//
//  Action Replay control program
//
//  Copyright (C) Foo Chen Hon 1997. All rights reserved.
//

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <io.h>
#include <fcntl.h>


typedef unsigned char	uchar;
typedef unsigned short	ushort;
typedef unsigned long	ulong;


#define DEF_TIMEOUT	30000


void CheckForEscape()
{
    if (kbhit())
	if (getch() == 0x1B)
	{
	    printf("ABORTED!\n");
	    exit(-1);
	}
}


int SendAndCheck(uchar data, uchar check, ushort timeout)
{
	ushort	to = timeout;

    outp(0x300, data);
    while (inp(0x302) & 0x01)
    {
	CheckForEscape();
	if (timeout)
	    if (!to--)
	    {
		to = timeout;
		printf("WARNING: Timeout - press [ESC] to quit!\n");
	    }
    }
    if ((data = inp(0x300)) == check)
	return 1;
    printf("ERROR: Expected - %02Xh  Received - %02Xh\n", check, data);
    exit(-1);
}


void WaitForSystem(uchar data, uchar check)
{
    printf("Waiting for PlayStation...");
    do
    {
	outp(0x300, data);
	while (inp(0x302) & 0x01)
	    CheckForEscape();
	CheckForEscape();
    } while (inp(0x300) != check);
    printf("READY!\n");
}


void SendLong(long lData)
{
	uchar	data;

    data = (uchar)((lData >> 24) & 0x000000FF);
    SendAndCheck(data, data, DEF_TIMEOUT);
    data = (uchar)((lData >> 16) & 0x000000FF);
    SendAndCheck(data, data, DEF_TIMEOUT);
    data = (uchar)((lData >> 8) & 0x000000FF);
    SendAndCheck(data, data, DEF_TIMEOUT);
    data = (uchar)(lData & 0x000000FF);
    SendAndCheck(data, data, DEF_TIMEOUT);
}


void SendFileData(int fh, ulong lLength)
{
	static uchar	buffer[4096];
	ushort		i, nread, csum = 0;

    while (lLength)
    {
	nread = (lLength >= 4096) ? 4096 : (ushort)lLength;
	if ((nread = read(fh, buffer, nread)) == -1)
	{
	    printf("ERROR: Reading file!\n");
	    close(fh);
	    exit(-1);
	}
	lLength -= nread;
	for (i = 0; i < nread; i++)
	{
	    SendAndCheck(buffer[i], buffer[i], DEF_TIMEOUT);
	    csum = (csum + buffer[i]) & 0x0FFF;
	}
    }
    SendAndCheck((uchar)(csum >> 8), (uchar)(csum >> 8), DEF_TIMEOUT);
    SendAndCheck((uchar)(csum & 0x00FF), (uchar)(csum & 0x00FF), DEF_TIMEOUT);
}


void ExecuteAddress(ulong start)
{
    printf("Execute at address %08lX...\n", start);
    WaitForSystem('W', 'R');
    SendAndCheck('B', 'W', DEF_TIMEOUT);
    SendAndCheck('X', 'X', DEF_TIMEOUT);
    SendLong(start);
    SendLong(0);
    SendAndCheck(0, 0, DEF_TIMEOUT);
    SendAndCheck(0, 0, DEF_TIMEOUT);
    SendAndCheck(0, 'O', DEF_TIMEOUT);
    SendAndCheck(0, 'K', DEF_TIMEOUT);
    printf("Command sent!\n");
}


void DownloadProgram(int fh, ulong start, ulong length)
{
    printf("Downloading program image...\n");
    WaitForSystem('W', 'R');
    SendAndCheck('B', 'W', DEF_TIMEOUT);
    SendAndCheck('X', 'X', DEF_TIMEOUT);
    SendLong(start);
    SendLong(length);
    SendFileData(fh, length);
    SendAndCheck(0, 'O', DEF_TIMEOUT);
    SendAndCheck(0, 'K', DEF_TIMEOUT);
    printf("Program downloaded!\n");
}


void UpgradeROM(int fh, ulong length)
{
	int	flh;
	ulong	flen;

    printf("Loading FLASHER...\n");
    if ((flh = open("FLASHER.ROM", O_RDONLY | O_BINARY)) == -1)
    {
	printf("ERROR: Opening FLASHER.ROM file!\n");
	exit(-1);
    }
    flen = filelength(flh);
    WaitForSystem('W', 'R');
    SendAndCheck('B', 'W', DEF_TIMEOUT);
    SendAndCheck('X', 'X', DEF_TIMEOUT);
    SendLong(0x80020000);
    SendLong(flen);
    SendFileData(flh, flen);
    SendAndCheck(0, 'O', DEF_TIMEOUT);
    SendAndCheck('W', 'K', DEF_TIMEOUT);
    close(flh);
    printf("FLASHER file loaded!\n");

    printf("Upgrading ROM image...\n");
    WaitForSystem('W', 'R');
    SendAndCheck('B', 'W', DEF_TIMEOUT);
    SendAndCheck('U', 'U', DEF_TIMEOUT);
    SendLong(0x80040000);
    SendLong(length);
    SendFileData(fh, length);
    SendAndCheck(0, 'O', DEF_TIMEOUT);
    SendAndCheck(0, 'K', DEF_TIMEOUT);
    printf("ROM upgraded!\n");
}


void main(int argc, char *argv[])
{
	char	command = 'X', *fname = NULL;
	ulong	start = 0x80010000, offset = 0, length = 0, flen, execute = 0;
	int	fh;

    printf("Action Replay Control             Version 0.01\n");
    printf("Copyright (C) Foo Chen Hon 1997.  All rights reserved.\n");

    // check for command line parameters
    while (--argc)
    {
	if (argv[argc][0] != '/')
	    fname = argv[argc];
	else switch (argv[argc][1])
	{
	    case 'S':			// start address
	    case 's':
		sscanf(argv[argc] + (argv[argc][2] == ':' ? 3 : 2), "%lX",
		    &start);
		break;

	    case 'O':			// file offset
	    case 'o':
		sscanf(argv[argc] + (argv[argc][2] == ':' ? 3 : 2), "%lX",
		    &offset);
		break;

	    case 'L':			// length
	    case 'l':
		sscanf(argv[argc] + (argv[argc][2] == ':' ? 3 : 2), "%lX",
		    &length);
		break;

	    case 'X':			// load & execute
	    case 'x':
		command = 'X';
		break;

	    case 'U':			// load & upgrade ROM image
	    case 'u':
		command = 'U';
		break;

	    case 'G':			// file offset
	    case 'g':
		sscanf(argv[argc] + (argv[argc][2] == ':' ? 3 : 2), "%lX",
		    &execute);
		ExecuteAddress(execute);
		exit(0);
		break;
	}
    }
    if (!fname)
    {
	printf("Usage: ARCMD <file> [/S<n>] [/O<n>] [/L<n>] [/G<n>] [/X] [/U]\n");
	printf("       /S<n>    load file to address <n> on PSX\n");
	printf("       /O<n>    load file from offset <n>\n");
	printf("       /L<n>    load <n> bytes from file\n");
	printf("       /O<n>    load file from offset <n>\n");
	printf("       /G<n>    go to address <n> & execute\n");
	printf("       /X       load & execute\n");
	printf("       /U       upgrade ROM image\n");
	exit(-1);
    }

    if ((fh = open(fname, O_RDONLY | O_BINARY)) == -1)
    {
	printf("ERROR: Opening file %s!\n", fname);
	exit(-1);
    }
    if (offset)
	lseek(fh, offset, SEEK_SET);
    flen = filelength(fh);
    if (!length)
	length = flen;
    else if ((length + offset) > flen)
	length = flen - offset;

    if (command == 'X')
	DownloadProgram(fh, start, length);
    else if (command == 'U')
	UpgradeROM(fh, length);

    close(fh);
    exit(0);
}
